#include <unistd.h>
#include <string.h>
#include <stdio.h>
#include <stdlib.h>
#include <fcntl.h>
#include <sys/stat.h>
#include "libifkoi.h"

const int ifkoi[128]={0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 2, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 2, 0, 48, 5409, 565, 327, 1854, 6380, 431, 1139, 942, 6642, 900, 2626, 2524, 1524, 5438, 6688, 1361, 873, 3244, 3545, 4066, 1484, 358, 2756, 718, 1530, 965, 315, 223, 92, 1229, -3, -5427, -672, -2740, -1176, -1854, -6380, -280, -805, -6603, -855, -2662, -2571, -1760, -5392, -6659, -1450, -3205, -3783, -3888, -1159, -253, -911, -462, -944, -324, -152, 36, -1518, -671, -213, -224, -1236};

char fromwin[] = {0xb3, 0xa9, 0xaa, 0xab, 0xac, 0xad, 0xae, 0xaf, 0xb0, 0xb1, 0xb2, 0xb3, 0xb4, 0xb5, 0xb6, 0xb7, 0xa3, 0xb9, 0xba, 0xbb, 0xbc, 0xbd, 0xbe, 0xbf, 0xe1, 0xe2, 0xf7, 0xe7, 0xe4, 0xe5, 0xf6, 0xfa, 0xe9, 0xea, 0xeb, 0xec, 0xed, 0xee, 0xef, 0xf0, 0xf2, 0xf3, 0xf4, 0xf5, 0xe6, 0xe8, 0xe3, 0xfe, 0xfb, 0xfd, 0xff, 0xf9, 0xf8, 0xfc, 0xe0, 0xf1, 0xc1, 0xc2, 0xd7, 0xc7, 0xc4, 0xc5, 0xd6, 0xda, 0xc9, 0xca, 0xcb, 0xcc, 0xcd, 0xce, 0xcf, 0xd0, 0xd2, 0xd3, 0xd4, 0xd5, 0xc6, 0xc8, 0xc3, 0xde, 0xdb, 0xdd, 0xdf, 0xd9, 0xd8, 0xdc, 0xc0, 0xd1};

char towin[] = {0xb8, 0xa4, 0xa5, 0xa6, 0xa7, 0xa8, 0xa9, 0xaa, 0xab, 0xac, 0xad, 0xae, 0xaf, 0xb0, 0xb1, 0xb2, 0xa8, 0xb4, 0xb5, 0xb6, 0xb7, 0xb8, 0xb9, 0xba, 0xbb, 0xbc, 0xbd, 0xbe, 0xbf, 0xfe, 0xe0, 0xe1, 0xf6, 0xe4, 0xe5, 0xf4, 0xe3, 0xf5, 0xe8, 0xe9, 0xea, 0xeb, 0xec, 0xed, 0xee, 0xef, 0xff, 0xf0, 0xf1, 0xf2, 0xf3, 0xe6, 0xe2, 0xfc, 0xfb, 0xe7, 0xf8, 0xfd, 0xf9, 0xf7, 0xfa, 0xde, 0xc0, 0xc1, 0xd6, 0xc4, 0xc5, 0xd4, 0xc3, 0xd5, 0xc8, 0xc9, 0xca, 0xcb, 0xcc, 0xcd, 0xce, 0xcf, 0xdf, 0xd0, 0xd1, 0xd2, 0xd3, 0xc6, 0xc2, 0xdc, 0xdb, 0xc7, 0xd8, 0xdd, 0xd9, 0xd7, 0xda};

char buf[4096];
char tmpfname[2048];
struct stat sb;

int removemeta=0;

int rmmeta(char *buf, int size) {
	int i;
	char *p=buf-1;
	int inside=0;

	while((++p - buf) < size) {
		if(*p=='>') { inside=0; continue; }
		if(*p == '<') { inside=1; continue; }
		if(inside) {
			/* forwarding to a first non-space character */
			while((p-buf) < size) if(*p != ' ') break; else p++;
			  if((p-buf) >= size) return 0;
			if((*p != 'm') && (*p != 'M')) { inside=0; continue; };
			  p++; if((p - buf) >= size) return 0;
			if((*p != 'e') && (*p != 'E')) { inside=0; continue; }
			  p++; if((p - buf) >= size) return 0;
			if((*p != 't') && (*p != 'T')) { inside=0; continue; }
			  p++; if((p - buf) >= size) return 0;
			if((*p != 'a') && (*p != 'A')) { inside=0; continue; }
			  p++; if((p - buf) >= size) return 0;
			if(*p != ' ') { inside=0; continue; }
			*(p-4)='t';
			*(p-2)='m';
		}
	}
	return 0;
}

int ifkoi_file(const char *file) {
	int balance=0;
	int size;
	unsigned char *p;
	int fd=open(file, O_RDONLY);

	if(fd==-1) return -2;
	if(fstat(fd, &sb) || ((sb.st_mode & S_IFMT) != S_IFREG)) return -2;

	while(size=read(fd, &buf, sizeof(buf))) {
		p=(unsigned char *)buf;
		while(size--) {
			if(*p > 127)
				balance+=ifkoi[*p-128];
			else
				if(*p < 10) return -1;
			p++;
		}
		if(abs(balance) > 1000000) break;
	}

	if(!balance) return -1;
	return (balance>0);
}

int ifkoi_av(char **av) {
	char **word=av;
	unsigned char *sym;
	int balance=0;

	while(*word) {
		sym=*word;
		while(*sym) {
			if(*sym > 127)
				balance+=ifkoi[*sym-128];
			else
				if(*sym < 10) return -1;
			sym++;
		}
		if(abs(balance) > 1000000) break;
		word++;
	}

	if(!balance) return -1;
	return (balance>0);
}

int tokoi_av(char **av) {
	char **word=av;
	char *sym;

	if(ifkoi_av(av)) return 1;

	while(*word) {
		sym=*word;
		while(*sym) {
			if( *(unsigned char *)sym >= 0xa8 )
				*sym=fromwin[ *(unsigned char *)sym - 0xa8 ];
			sym++;
		}
		word++;
	}

	return 0;
}

int tokoi_file(const char *fname) {
	int fd, out, size, sizen;
	int removed=0;
	char *p;

	if((size=ifkoi_file(fname)) < 0) return -1;
	if((fd=open(fname, O_RDONLY)) == -1) return -1;
	strncpy(tmpfname, fname, sizeof(tmpfname)-10);
	strcat(tmpfname, "XXXXXX");
	if((out=mkstemp(tmpfname)) == -1) return -1;

	if(size) {
		while(size=read(fd, &buf, sizeof(buf))) {
			if(removemeta && !(removed++)) rmmeta(buf, size);
			write(out, &buf, size);
		}
	} else {
		while(size=sizen=read(fd, &buf, sizeof(buf))) {
			if(removemeta && !(removed++)) rmmeta(buf, size);
			p=buf;
			while(size--) {
				if( *(unsigned char *)p >= 0xa8 )
				*p=fromwin[ *(unsigned char *)p - 0xa8 ];
				p++;
			}
			write(out, &buf, sizen);
		}
	}
	close(fd);

	fchown(out, sb.st_uid, sb.st_gid);
	fchmod(out, sb.st_mode & 0777);
	close(out);

	if(rename(tmpfname, fname)) {
		fprintf(stderr, "Can't rename()\n");
		unlink (tmpfname);
	}

	return 0;
}

int towin_av(char **av) {
	char **word=av;
	char *sym;

	if(ifkoi_av(av)<=0) return 1;

	while(*word) {
		sym=*word;
		while(*sym) {
			if( *(unsigned char *)sym >= 0xa3 )
				*sym=towin[ *(unsigned char *)sym - 0xa3 ];
			sym++;
		}
		word++;
	}

	return 0;
}

int towin_file(const char *fname) {
	int fd, out, size, sizen;
	int removed=0;
	char *p;

	if((size=ifkoi_file(fname)) < 0) return -1;
	if((fd=open(fname, O_RDONLY)) == -1) return 1;
	strncpy(tmpfname, fname, sizeof(tmpfname)-10);
	strcat(tmpfname, "XXXXXX");
	if((out=mkstemp(tmpfname)) == -1) return -1;

	if(size==0) {
		while(size=read(fd, &buf, sizeof(buf))) {
			if(removemeta && !(removed++)) rmmeta(buf, size);
			write(out, &buf, size);
		}
	} else {
		while(size=sizen=read(fd, &buf, sizeof(buf))) {
			if(removemeta && !(removed++)) rmmeta(buf, size);
			p=buf;
			while(size--) {
				if( *(unsigned char *)p >= 0xa3 )
				*p=towin[ *(unsigned char *)p - 0xa3 ];
				p++;
			}
			write(out, &buf, sizen);
		}
	}
	close(fd);

	fchown(out, sb.st_uid, sb.st_gid);
	fchmod(out, sb.st_mode & 0777);
	close(out);

	if(rename(tmpfname, fname)) {
		fprintf(stderr, "Can't rename()\n");
		unlink (tmpfname);
	}

	return 0;
}

